#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/sysinfo.h>
#include <sys/statvfs.h>
#include <dirent.h>
#include <libgen.h>

#define CONFIG_FILE "conf.cfg"
uint16_t PORT = 8888; // defailt port
char PASSWORD[30];

#ifdef NDEBUG
#define DEBUG(x...) {}
#else
#define DEBUG(x...) {printf(x);fflush(stdout);}
#endif

int cpu_utilization = 0;

void read_config();

char* get_cpu_utilization();
char* get_free_memory();
char* get_disk_space();
char* get_process_running(const char* proc);

void *cpu_utilization_thread(void*);

int main(void)
{
    //default password
    snprintf(PASSWORD, sizeof(PASSWORD), "pass123");

    read_config();

    pthread_t th;
    pthread_create(&th, NULL, cpu_utilization_thread, NULL);

    int s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (s < 0) {
        fprintf(stderr, "Error create socket: %s(%d)\n", strerror(errno), errno);
        return s;
    }

    struct sockaddr_in si_server;
    bzero(&si_server, sizeof(si_server));

    si_server.sin_family = AF_INET;
    si_server.sin_port = htons(PORT);
    si_server.sin_addr.s_addr = htonl(INADDR_ANY);

    if( bind(s, (struct sockaddr*)&si_server, sizeof(si_server) ) == -1) {
        fprintf(stderr, "Error bind socket: %s(%d)\n", strerror(errno), errno);
        return -1;
    }

    int pass_len = strlen(PASSWORD);

    while(1) {
        struct sockaddr_in si_client;
        int slen = sizeof(si_client);
        int len;
        char buf[2048];
        if ((len = recvfrom(s, buf, sizeof(buf)-1, 0, (struct sockaddr *) &si_client, (socklen_t *) &slen)) < -1) {
            fprintf(stderr, "Error recvfrom: %s(%d)\n", strerror(errno), errno);
            return -1;
        }
        buf[len] = 0;

        DEBUG("got line: %s\n", buf);

        // input must be greater than password len and delimeter "::"
        if (len < pass_len + 2) {
            DEBUG("input too short\n");
            continue;
        }

        // check delimeter "::" on place and password is correct
        if (buf[pass_len] != ':' || buf[pass_len+1] != ':' || strncmp(buf,PASSWORD, pass_len)) {
            DEBUG("incorrect password: %s\n", buf);
            continue;
        }

        //looking for password (pass123::CPU)
        char * pass = strstr(buf, "::");
        if (!pass) {
            DEBUG("password not found\n");
            continue;
        }
        if (pass - buf != pass_len || strncmp(buf,PASSWORD,pass_len) ) {
            *pass = 0;
            DEBUG("incorrect password: %s\n", buf);
            continue;
        }

        // calculate pointer to command: skip password and delimeter "::"
        char* cmd = pass + 2;

        char * ret = NULL;

        if (!strcmp(cmd, "CPU")) {
            ret = get_cpu_utilization();
        } else if (!strcmp(cmd, "MEM")) {
            ret = get_free_memory();
        } else if (!strcmp(cmd, "DISK-SPACE")) {
            ret = get_disk_space();
        } else if (!strncmp(cmd, "PROCESS::", strlen("PROCESS::"))) {
            ret = get_process_running(cmd + strlen("PROCESS::"));
        } else {
            DEBUG("incorrect command\n");
            continue;
        }

        if (!ret) {
            fprintf(stderr, "command process failed: %s\n", cmd);
            continue;
        }

        if (sendto(s, ret, strlen(ret), 0, (struct sockaddr*) &si_client, slen) < 0) {
            fprintf(stderr, "sendto failed: %s(%d)\n", strerror(errno), errno);
            return -1;
        }
        DEBUG("sent answer: %s(%lu)\n", ret, strlen(ret));
    }

    close(s);

    if(pthread_join(th, NULL)) {
        fprintf(stderr, "Error joining thread\n");
        return -1;
    }

    return 0;
}

void read_config() {
    FILE* f = fopen(CONFIG_FILE, "r");
    if (!f) {
        fprintf(stderr, "Config file %s not found\n", CONFIG_FILE);
        return;
    }
    char rbuf[80];
    while(fgets(rbuf, sizeof(rbuf), f)) {
        char *key = strtok(rbuf,":");
        char *value = strtok(NULL, ":");
        if (!key || !value) {
            //fprintf(stderr, "Error parse config file line: %s\n", rbuf);
            continue;
        }
        if (!strcmp(key, "PORT")) {
            PORT = strtoul(value, NULL, 10);
        }
        if (!strcmp(key, "PASSWORD")) {
            while(*value == ' ') value++; //strip spaces before pass
            while(
                    value[strlen(value)-1] == ' ' ||
                    value[strlen(value)-1] == '\n' || 
                    value[strlen(value)-1] == '\r' ||
                    value[strlen(value)-1] == '\t'
                    ) value[strlen(value)-1] = 0; //strip spaces after pass
            snprintf(PASSWORD, sizeof(PASSWORD), "%s", value);
        }
    }
    fclose(f);
}

char* get_cpu_utilization() {
    static char buf[10] = { 0 };
    snprintf(buf, sizeof(buf), "%d", cpu_utilization);
    return buf;
}

char* get_free_memory() {
    static char buf[4];
    buf[0] = 0;

    FILE* f = fopen("/proc/meminfo", "r");
    if (!f) {
        fprintf(stderr, "Error get meminfo\n");
        return buf;
    }
    char rbuf[80];
    unsigned long mem_total = 0;
    unsigned long mem_avail = 0;
    while(fgets(rbuf, sizeof(rbuf), f)) {
        char *key = strtok(rbuf,":");
        char *value = strtok(NULL, ":");
        if (!key || !value) {
            fprintf(stderr, "Error parse meminfo\n");
            fclose(f);
            return buf;
        }
        if (!strcmp(key, "MemTotal")) {
            mem_total = strtoul(value, NULL, 10);
        }
        if (!strcmp(key, "MemAvailable")) {
            mem_avail = strtoul(value, NULL, 10);
        }
        if (mem_total && mem_avail) {
            int mem = (int)(1.0 * mem_avail / mem_total * 100.0);
            snprintf(buf, sizeof(buf), "%d", mem);
            fclose(f);
            return buf;
        }
    }
    fclose(f);
    return buf;
}

char* get_disk_space() {
    static char buf[1024];
    buf[0] = 0;
    FILE* f = fopen("/proc/mounts", "r");
    if (!f) {
        fprintf(stderr, "Error get mounts\n");
        return buf;
    }
    char rbuf[512];
    int disk = 1;
    while(fgets(rbuf, sizeof(rbuf), f)) {
        if (*rbuf != '/') continue;
        char *key = strtok(rbuf," ");
        char *value = strtok(NULL, " ");
        if (!key || !value) {
            fprintf(stderr, "Error parse mounts\n");
            fclose(f);
            return buf;
        }

        struct statvfs stat;
        if (statvfs(value, &stat) < 0) {
            fprintf(stderr, "Error statvfs\n");
            fclose(f);
            return buf;
        }
        int disk_free = (int)(1.0 * stat.f_bavail / stat.f_blocks * 100.0);
        if (strlen(buf) && strlen(buf) < sizeof(buf)-1) {
            strcat(buf, "\n");
        }
        char buf2[50];
        snprintf(buf2, sizeof(buf2), "disk%d:%d", disk, disk_free);
        if (strlen(buf) + strlen(buf2) < sizeof(buf)-1) {
            strcat(buf, buf2);
        }
        ++disk;
    }

    fclose(f);
    return buf;
}

char* get_process_running(const char* proc) {
    static char buf[2];
    buf[0] = 0;

    DIR* dir;
    if (!(dir = opendir("/proc"))) {
        perror("can't open /proc");
        return buf;
    }

    struct dirent* ent;
    char* endptr;
    while((ent = readdir(dir)) != NULL) {
        long lpid = strtol(ent->d_name, &endptr, 10);
        if (*endptr != '\0') {
            continue;
        }

        char rbuf[512];
        snprintf(rbuf, sizeof(rbuf), "/proc/%ld/cmdline", lpid);
        FILE* fp = fopen(rbuf, "r");

        if (fp) {
            if (fgets(rbuf, sizeof(rbuf), fp) != NULL) {
                char* first = strtok(rbuf, " ");
                if (!strcmp(basename(first), proc)) {
                    fclose(fp);
                    closedir(dir);
                    buf[0] = '1';
                    return buf;
                }
            }
            fclose(fp);
        }

    }
    closedir(dir);

    buf[0] = '2';
    return buf;
}

void *cpu_utilization_thread(void* arg) {
    while (1) {
        long double p1[4], p2[4];
        FILE* f = fopen("/proc/stat","r");
        if (fscanf(f,"%*s %Lf %Lf %Lf %Lf",&p1[0],&p1[1],&p1[2],&p1[3]) != 4) {
            DEBUG("cpu_utilization_thread fscanf1 failed\n");
        }
        fclose(f);
        sleep(1);

        f = fopen("/proc/stat","r");
        if (fscanf(f,"%*s %Lf %Lf %Lf %Lf",&p2[0],&p2[1],&p2[2],&p2[3]) != 4) {
            DEBUG("cpu_utilization_thread fscanf2 failed\n");
        }
        fclose(f);

        cpu_utilization = (int)((((p2[0]+p2[1]+p2[2]) - (p1[0]+p1[1]+p1[2])) / ((p2[0]+p2[1]+p2[2]+p2[3]) - (p1[0]+p1[1]+p1[2]+p1[3])))*100);
    }
}

